/* Copyright (C) 2013-2014 Patrick Brans.  All rights reserved.
**
** This file is part of BrickStock.
** BrickStock is based heavily on BrickStore (http://www.brickforge.de/software/brickstore/)
** by Robert Griebl, Copyright (C) 2004-2008.
**
** This file may be distributed and/or modified under the terms of the GNU 
** General Public License version 2 as published by the Free Software Foundation 
** and appearing in the file LICENSE.GPL included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://fsf.org/licensing/licenses/gpl.html for GPL licensing information.
*/
#ifndef __CUPDATEDATABASE_H__
#define __CUPDATEDATABASE_H__

#include <qfile.h>
#include <qdatetime.h>

#include "lzmadec.h"

#include "version.h"
#include "cconfig.h"
#include "bricklink.h"
#include "cprogressdialog.h"

class CUpdateDatabase : public QObject {
	Q_OBJECT

public:
	CUpdateDatabase ( CProgressDialog *pd )
		: m_progress ( pd )
	{
		connect ( pd, SIGNAL( transferFinished ( )), this, SLOT( gotten ( )));

		pd-> setAutoClose ( false );
		pd-> setHeaderText ( tr( "Updating BrickLink Database" ));
		pd-> setMessageText ( tr( "Download: %1/%2 KB" ));

        QString remotefile = "http://" BRICKSTOCK_URL BRICKSTOCK_DB_PATH + BrickLink::inst ( )-> defaultDatabaseName ( );
		QString localfile = BrickLink::inst ( )-> dataPath ( ) + BrickLink::inst ( )-> defaultDatabaseName ( );

		QDateTime dt;
		if ( QFile::exists ( localfile ))
			dt = CConfig::inst ( )-> lastDatabaseUpdate ( );

		QFile *file = new QFile ( localfile + ".lzma" );

		if ( file-> open ( QIODevice::WriteOnly )) {
			pd-> get ( remotefile + ".lzma", CKeyValueList ( ), dt, file );
		}
		else {
			pd-> setErrorText ( tr( "Could not write to file: %1" ). arg( file-> name ( )));
			delete file;
		}
	}

private slots:
	virtual void gotten ( )
	{
		CTransfer::Job *job = m_progress-> job ( );
		QFile *file = job-> file ( );

		if ( job-> notModifiedSince ( )) {
			file-> remove ( );
			m_progress-> setMessageText ( tr( "Already up-to-date." ));
			m_progress-> setFinished ( true );
		}
		else if ( file-> size ( )) {
			QString basepath = file-> name ( );
			basepath. truncate ( basepath. length ( ) - 5 ); // strip '.lzma'

			QString error = decompress ( file-> name ( ), basepath );

			if ( error. isNull ( )) {
				if ( BrickLink::inst ( )-> readDatabase ( )) {
                    CConfig::inst ( )-> setLastDatabaseUpdate ( QDateTime::currentDateTime ( ));

					m_progress-> setMessageText ( tr( "Finished." ));
					m_progress-> setFinished ( true );
				}
				else
					m_progress-> setErrorText ( tr( "Could not load the new database." ));
			}
			else
				m_progress-> setErrorText ( error );
		}
		else
			m_progress-> setErrorText ( tr( "Downloaded file is empty." ));
	}

private:
	QString decompress ( const QString &src, const QString &dst )
	{
		QFile sf ( src );
		QFile df ( dst );
		
		if ( !sf. open ( QIODevice::ReadOnly ))
			return tr( "Could not read downloaded file: %1" ). arg( src );
		if ( !df. open ( QIODevice::WriteOnly ))
			return tr( "Could not write to database file: %1" ). arg( dst );
		
		static const int CHUNKSIZE_IN = 4096;
		static const int CHUNKSIZE_OUT = 512 * 1024;
		
		char *buffer_in  = new char [CHUNKSIZE_IN];
		char *buffer_out = new char [CHUNKSIZE_OUT];

		lzmadec_stream strm;
		strm. lzma_alloc = 0;
		strm. lzma_free = 0;
		strm. opaque = 0;
		strm. avail_in = 0;
		strm. next_in = 0;
		
		if ( lzmadec_init ( &strm ) != LZMADEC_OK )
			return tr( "Could not initialize the LZMA decompressor" );


		QString loop_error;
		
		m_progress-> setMessageText ( tr( "Decompressing database" ));
		m_progress-> setProgress ( 0, 0 );

		while ( true ) {
			if ( strm. avail_in == 0 ) {
				strm. next_in  = (unsigned char *) buffer_in;
				strm. avail_in = sf. readBlock ( buffer_in, CHUNKSIZE_IN );
			}
			strm. next_out  = (unsigned char *) buffer_out;
			strm. avail_out = CHUNKSIZE_OUT;
			
			int ret = lzmadec_decode ( &strm, strm.avail_in == 0 );
			if ( ret != LZMADEC_OK && ret != LZMADEC_STREAM_END ) {
				loop_error = tr( "Error while decompressing %1" ). arg( src );
				break;
			}
				
			Q_LONG write_size = CHUNKSIZE_OUT - strm.avail_out;
			if ( write_size != df. writeBlock ( buffer_out, write_size )) {
				loop_error = tr( "Error writing to file %1: %2" ). arg( dst, df. errorString ( ));
				break;
			}
			if ( ret == LZMADEC_STREAM_END ) {
				lzmadec_end ( &strm );
				break;
			}
			m_progress-> setProgress ( sf. at ( ), sf. size ( ));
		}
		
		delete [] buffer_in;
		delete [] buffer_out;
		
		return loop_error;
	}


private:
	CProgressDialog *m_progress;
};

#endif
